# Copyright 2017 Capital One Services, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
SQS Message Processing
===============

"""
import base64
import json
import logging
import traceback
import zlib

import six

from .email_delivery import EmailDelivery
from .sns_delivery import SnsDelivery

from c7n_mailer.utils import kms_decrypt

DATA_MESSAGE = "maidmsg/1.0"


class MailerSqsQueueIterator(object):
    # Copied from custodian to avoid runtime library dependency
    msg_attributes = ['sequence_id', 'op', 'ser']

    def __init__(self, aws_sqs, queue_url, logger, limit=0, timeout=10):
        self.aws_sqs = aws_sqs
        self.queue_url = queue_url
        self.limit = limit
        self.logger = logger
        self.timeout = timeout
        self.messages = []

    # this and the next function make this object iterable with a for loop
    def __iter__(self):
        return self

    def __next__(self):
        if self.messages:
            return self.messages.pop(0)
        response = self.aws_sqs.receive_message(
            QueueUrl=self.queue_url,
            WaitTimeSeconds=self.timeout,
            MaxNumberOfMessages=3,
            MessageAttributeNames=self.msg_attributes,
            AttributeNames=['SentTimestamp']
        )

        msgs = response.get('Messages', [])
        self.logger.debug('Messages received %d', len(msgs))
        for m in msgs:
            self.messages.append(m)
        if self.messages:
            return self.messages.pop(0)
        raise StopIteration()

    next = __next__  # python2.7

    def ack(self, m):
        self.aws_sqs.delete_message(
            QueueUrl=self.queue_url,
            ReceiptHandle=m['ReceiptHandle'])


class MailerSqsQueueProcessor(object):

    def __init__(self, config, session, logger, max_num_processes=16):
        self.config = config
        self.logger = logger
        self.session = session
        self.max_num_processes = max_num_processes
        self.receive_queue = self.config['queue_url']
        if self.config.get('debug', False):
            self.logger.debug('debug logging is turned on from mailer config file.')
            logger.setLevel(logging.DEBUG)

    """
    Cases
    - aws resource is tagged CreatorName: 'milton', ldap_tag_uids has CreatorName,
        we do an ldap lookup, get milton's email and send him an email
    - you put an email in the to: field of the notify of your policy, we send an email
        for all resources enforce by that policy
    - you put an sns topic in the to: field of the notify of your policy, we send an sns
        message for all resources enforce by that policy
    - an lambda enforces a policy based on an event, we lookup the event aws username, get their
        ldap email and send them an email about a policy enforcement (from lambda) for the event
    - resource-owners has a list of tags, SupportEmail, OwnerEmail, if your resources
        include those tags with valid emails, we'll send an email for those resources
        any others
    - resource-owners has a list of tags, SnSTopic, we'll deliver an sns message for
        any resources with SnSTopic set with a value that is a valid sns topic.
    """
    def run(self, parallel=False):
        self.logger.info("Downloading messages from the SQS queue.")
        aws_sqs = self.session.client('sqs')
        sqs_messages = MailerSqsQueueIterator(aws_sqs, self.receive_queue, self.logger)

        sqs_messages.msg_attributes = ['mtype', 'recipient']
        # lambda doesn't support multiprocessing, so we don't instantiate any mp stuff
        # unless it's being run from CLI on a normal system with SHM
        if parallel:
            import multiprocessing
            process_pool = multiprocessing.Pool(processes=self.max_num_processes)
        for sqs_message in sqs_messages:
            self.logger.debug(
                "Message id: %s received %s" % (
                    sqs_message['MessageId'], sqs_message.get('MessageAttributes', '')))
            msg_kind = sqs_message.get('MessageAttributes', {}).get('mtype')
            if msg_kind:
                msg_kind = msg_kind['StringValue']
            if not msg_kind == DATA_MESSAGE:
                warning_msg = 'Unknown sqs_message or sns format %s' % (sqs_message['Body'][:50])
                self.logger.warning(warning_msg)
            if parallel:
                process_pool.apply_async(self.process_sqs_message, args=sqs_message)
            else:
                self.process_sqs_message(sqs_message)
            self.logger.debug('Processed sqs_message')
            sqs_messages.ack(sqs_message)
        if parallel:
            process_pool.close()
            process_pool.join()
        self.logger.info('No sqs_messages left on the queue, exiting c7n_mailer.')
        return

    # This function when processing sqs messages will only deliver messages over email or sns
    # If you explicitly declare which tags are aws_usernames (synonymous with ldap uids)
    # in the ldap_uid_tags section of your mailer.yml, we'll do a lookup of those emails
    # (and their manager if that option is on) and also send emails there.
    def process_sqs_message(self, encoded_sqs_message):
        body = encoded_sqs_message['Body']
        try:
            body = json.dumps(json.loads(body)['Message'])
        except ValueError:
            pass
        sqs_message = json.loads(zlib.decompress(base64.b64decode(body)))

        self.logger.debug("Got account:%s message:%s %s:%d policy:%s recipients:%s" % (
            sqs_message.get('account', 'na'),
            encoded_sqs_message['MessageId'],
            sqs_message['policy']['resource'],
            len(sqs_message['resources']),
            sqs_message['policy']['name'],
            ', '.join(sqs_message['action'].get('to'))))

        # get the map of email_to_addresses to mimetext messages (with resources baked in)
        # and send any emails (to SES or SMTP) if there are email addresses found
        email_delivery = EmailDelivery(self.config, self.session, self.logger)
        to_addrs_to_email_messages_map = email_delivery.get_to_addrs_email_messages_map(sqs_message)
        for email_to_addrs, mimetext_msg in six.iteritems(to_addrs_to_email_messages_map):
            email_delivery.send_c7n_email(sqs_message, list(email_to_addrs), mimetext_msg)

        # this sections gets the map of sns_to_addresses to rendered_jinja messages
        # (with resources baked in) and delivers the message to each sns topic
        sns_delivery = SnsDelivery(self.config, self.session, self.logger)
        sns_message_packages = sns_delivery.get_sns_message_packages(sqs_message)
        sns_delivery.deliver_sns_messages(sns_message_packages, sqs_message)

        # this section sends a notification to the resource owner via Slack
        if any(e.startswith('slack') or e.startswith('https://hooks.slack.com/')
                for e in sqs_message.get('action', ()).get('to', []) +
                sqs_message.get('action', ()).get('owner_absent_contact', [])):
            from .slack_delivery import SlackDelivery

            if self.config.get('slack_token'):
                self.config['slack_token'] = \
                    kms_decrypt(self.config, self.logger, self.session, 'slack_token')

            slack_delivery = SlackDelivery(self.config, self.logger, email_delivery)
            slack_messages = slack_delivery.get_to_addrs_slack_messages_map(sqs_message)
            try:
                slack_delivery.slack_handler(sqs_message, slack_messages)
            except Exception:
                traceback.print_exc()
                pass

        # this section gets the map of metrics to send to datadog and delivers it
        if any(e.startswith('datadog') for e in sqs_message.get('action', ()).get('to')):
            from .datadog_delivery import DataDogDelivery
            datadog_delivery = DataDogDelivery(self.config, self.session, self.logger)
            datadog_message_packages = datadog_delivery.get_datadog_message_packages(sqs_message)

            try:
                datadog_delivery.deliver_datadog_messages(datadog_message_packages, sqs_message)
            except Exception:
                traceback.print_exc()
                pass

        # this section sends the full event to a Splunk HTTP Event Collector (HEC)
        if any(
            e.startswith('splunkhec://')
            for e in sqs_message.get('action', ()).get('to')
        ):
            from .splunk_delivery import SplunkHecDelivery
            splunk_delivery = SplunkHecDelivery(self.config, self.session, self.logger)
            splunk_messages = splunk_delivery.get_splunk_payloads(
                sqs_message, encoded_sqs_message['Attributes']['SentTimestamp']
            )

            try:
                splunk_delivery.deliver_splunk_messages(splunk_messages)
            except Exception:
                traceback.print_exc()
                pass



# Copyright 2019 Manheim / Cox Automotive
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
from time import sleep
from six.moves.urllib.parse import urlparse
from random import uniform
import requests
from jsonpointer import resolve_pointer, JsonPointerException
from jsonpatch import JsonPatch
from copy import deepcopy
from .utils import get_aws_username_from_event


class SplunkHecDelivery(object):
    """
    Delivery class to send c7n message from SQS to Splunk HTTP Event Collector
    """

    def __init__(self, config, session, logger):
        """
        Initialize SplunkHecDelivery HEC sender.

        :param config: sqs_splunk_lambda configuration
        :type config: dict
        :param session: boto3 AWS Session
        :param logger: Logger object to write to
        """
        self.config = config
        self.logger = logger
        self.session = session

    def get_splunk_payloads(self, msg, msg_timestamp):
        """
        Given a raw c7n message dict, generate a list of payloads to send to the
        Splunk HEC.

        :param msg: c7n notification message
        :type msg: dict
        :param msg_timestamp: int timestamp (milliseconds) when message was sent
          to SQS, for use as event timestamp in Splunk.
        :type msg_timestamp: int
        :return: list of Splunk payload dicts
        :rtype: list
        """
        payloads = []
        events = self.get_splunk_events(msg)
        indices = self._splunk_indices_for_message(msg)
        for event in events:
            for index in indices:
                payloads.append({
                    'time': msg_timestamp,
                    'host': 'cloud-custodian',
                    'source': '%s-cloud-custodian' % event.get(
                        'account', 'unknown'
                    ),
                    'sourcetype': '_json',
                    'index': index,
                    'event': event
                })
        return payloads

    def get_splunk_events(self, msg):
        """
        Given a raw c7n message dict, generate a list of per-resource event
        dictionaries to send to Splunk as events.

        :param msg: c7n notification message
        :type msg: dict
        :return: list of Splunk HEC Event dicts
        :rtype: list
        """
        # if an event is present, add it to the log
        if msg.get('event', None) is not None:
            # add user to the log message, at the top level...
            user = get_aws_username_from_event(self.logger, msg['event'])
            if user is not None:
                msg['event_triggering_user'] = user
        # get a copy of the message with no resources
        base_log = dict(msg)
        base_log.pop('resources')
        base_log = deepcopy(base_log)
        # if configured, build and add actions list
        if self.config.get('splunk_actions_list', False):
            base_log['actions'] = []
            for a in msg['policy']['actions']:
                if isinstance(a, type({})):
                    base_log['actions'].append(a['type'])
                else:
                    base_log['actions'].append(a)
        # generate a separate Splunk message for each resource
        logs = []
        for res in msg['resources']:
            x = dict(base_log)
            x['resource'] = dict(res)
            # ensure there's one "tags" element, and it's a dict
            tmp = self.tags_for_resource(res)
            if 'Tags' in x['resource']:
                del x['resource']['Tags']
            x['resource']['tags'] = tmp
            logs.append(self._prune_log_message(x))
        return logs

    def _prune_log_message(self, msg):
        """
        If the "splunk_remove_paths" config item is not set or empty, return
        ``msg`` unaltered. Otherwise, for each RFC6901 JSON Pointer in
        "splunk_remove_paths" that points to an element present in ``msg``,
        remove that element. Return the result.

        :param msg: Splunk-ready message
        :type msg: dict
        :return: msg dict with all ``splunk_remove_paths`` elements removed
        :rtype: dict
        """
        paths = self.config.get('splunk_remove_paths', [])
        if not paths:
            return msg
        patches = []
        for path in paths:
            try:
                resolve_pointer(msg, path)
                patches.append({'op': 'remove', 'path': path})
            except JsonPointerException:
                pass
        if not patches:
            return msg
        msg = JsonPatch(patches).apply(msg)
        return msg

    def deliver_splunk_messages(self, payloads):
        """
        Deliver log messages to Splunk,

        :param payloads: list of payload dicts to send to Splunk
        :type payloads: list
        """
        failed = 0
        for payload in payloads:
            if not self._try_send(payload):
                failed += 1
        if failed != 0:
            raise RuntimeError(
                'ERROR: {failed} of {count} Splunk HEC messages '
                'failed to deliver.'.format(
                    failed=failed, count=len(payloads)
                )
            )

    def _try_send(self, payload):
        """
        Retry sending payload to splunk via ``_send_splunk`` up to
        ``config["splunk_max_attempts"]`` times, sleeping a random amount
        of time between 1 and 4 seconds between each try.

        :param payload: the payload dict to send to Splunk as event data.
        :type payload: dict
        :return: True if sent successfully, False otherwise
        :rtype: bool
        """
        max_attempts = self.config.get('splunk_max_attempts', 4)
        maxlen = self.config.get('splunk_hec_max_length', None)
        p = json.dumps(payload)
        if maxlen is not None and len(p) > maxlen:
            # This is in place for Splunk installations that are configured
            # with a short maximum message length (i.e. 10,000 characters).
            self.logger.error(
                'ERROR: Sending %d characters to Splunk HEC; line length '
                'limit is %d characters. Data will be truncated: %s',
                len(p), maxlen, p
            )
        for i in range(0, max_attempts):
            try:
                self._send_splunk(p)
                if maxlen is not None and len(p) > maxlen:
                    # never retry if we're over configured max length
                    return False
                return True  # if no exception, just return
            except Exception:
                sleep_sec = uniform(1, 4)  # random float 1 to 4
                self.logger.warning(
                    'Caught exception sending to Splunk; '
                    'retry in %s seconds', sleep_sec
                )
                sleep(sleep_sec)
        self.logger.error(
            'ERROR - Could not POST to Splunk after %d tries.', max_attempts
        )
        return False

    def _send_splunk(self, payload):
        """
        Perform the actual data send to Splunk HEC for one log entry

        :param payload: the JSON-encoded payload to send to Splunk as
          event data
        :type payload: str
        """
        url = self.config['splunk_hec_url']
        self.logger.debug('Send to Splunk (%s): %s', url, payload)
        try:
            r = requests.post(
                url,
                headers={
                    'Authorization': 'Splunk %s' % self.config[
                        'splunk_hec_token'
                    ]
                },
                data=payload
            )
        except Exception:
            self.logger.error('Exception during Splunk POST to %s of %s',
                              url, payload, exc_info=True)
            raise
        self.logger.debug(
            'Splunk POST got response code %s HEADERS=%s BODY: %s',
            r.status_code, r.headers, r.text
        )
        if r.status_code not in [200, 201, 202]:
            self.logger.error(
                'Splunk POST returned non-20x response: %s HEADERS=%s BODY: %s',
                r.status_code, r.headers, r.text
            )
            raise RuntimeError('POST returned %s' % r.status_code)
        try:
            j = r.json()
        except Exception:
            j = {'text': r.text}
        if j['text'].lower() != 'success':
            self.logger.error(
                'Splunk POST returned non-success response: %s', j
            )
            raise RuntimeError('POST returned non-success response: %s' % j)

    def tags_for_resource(self, res):
        """
        Return the tags for a given resource, or an empty dict if they
        can't be found.

        :param res: c7n resource info
        :type res: dict
        :return: dict of tags
        """
        try:
            return {x['Key']: x['Value'] for x in res.get('Tags', [])}
        except Exception:
            self.logger.warning(
                'Exception building tags dict; Tags=%s',
                res.get('Tags', None)
            )
            return {}

    @staticmethod
    def _splunk_indices_for_message(msg):
        """
        Given a message body from c7n, return a list of splunk indices to send
        notifications to.

        :param msg: c7n notification message
        :type msg: dict
        :return: list of string Splunk index names
        :rtype: list
        """
        indices = set()
        if msg and msg.get('action', False) and msg['action'].get('to', False):
            for to in msg['action']['to']:
                if not to.startswith('splunkhec://'):
                    continue
                parsed = urlparse(to)
                indices.add(parsed.netloc)
        return sorted(list(indices))



# Copyright 2017 Capital One Services, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json

import re
import redis

try:
    import sqlite3
except ImportError:
    have_sqlite = False
else:
    have_sqlite = True
from ldap3 import Connection
from ldap3.core.exceptions import LDAPSocketOpenError


class LdapLookup(object):

    def __init__(self, config, logger):
        self.log = logger
        self.connection = self.get_connection(
            config.get('ldap_uri'),
            config.get('ldap_bind_user', None),
            config.get('ldap_bind_password', None)
        )
        self.base_dn = config.get('ldap_bind_dn')
        self.email_key = config.get('ldap_email_key', 'mail')
        self.manager_attr = config.get('ldap_manager_attribute', 'manager')
        self.uid_key = config.get('ldap_uid_attribute', 'sAMAccountName')
        self.attributes = ['displayName', self.uid_key, self.email_key, self.manager_attr]
        self.uid_regex = config.get('ldap_uid_regex', None)
        self.cache_engine = config.get('cache_engine', None)
        if self.cache_engine == 'redis':
            redis_host = config.get('redis_host')
            redis_port = int(config.get('redis_port', 6379))
            self.caching = self.get_redis_connection(redis_host, redis_port)
        elif self.cache_engine == 'sqlite':
            if not have_sqlite:
                raise RuntimeError('No sqlite available: stackoverflow.com/q/44058239')
            self.caching = LocalSqlite(config.get('ldap_cache_file', '/var/tmp/ldap.cache'), logger)

    def get_redis_connection(self, redis_host, redis_port):
        return Redis(redis_host=redis_host, redis_port=redis_port, db=0)

    def get_connection(self, ldap_uri, ldap_bind_user, ldap_bind_password):
        # note, if ldap_bind_user and ldap_bind_password are None
        # an anonymous bind will be attempted.
        try:
            return Connection(
                ldap_uri, user=ldap_bind_user, password=ldap_bind_password,
                auto_bind=True,
                receive_timeout=30,
                auto_referrals=False,
            )
        except LDAPSocketOpenError:
            self.log.error('Not able to establish a connection with LDAP.')

    def search_ldap(self, base_dn, ldap_filter, attributes):
        self.connection.search(base_dn, ldap_filter, attributes=self.attributes)
        if len(self.connection.entries) == 0:
            self.log.warning("user not found. base_dn: %s filter: %s", base_dn, ldap_filter)
            return {}
        if len(self.connection.entries) > 1:
            self.log.warning("too many results for search %s", ldap_filter)
            return {}
        return self.connection.entries[0]

    def get_email_to_addrs_from_uid(self, uid, manager=False):
        to_addrs = []
        uid_metadata = self.get_metadata_from_uid(uid)
        uid_email = uid_metadata.get(self.email_key, None)
        if uid_email:
            to_addrs.append(uid_email)
        if manager:
            uid_manager_dn = uid_metadata.get(self.manager_attr, None)
            uid_manager_email = None
            if uid_manager_dn:
                uid_manager = self.get_metadata_from_dn(uid_manager_dn)
                uid_manager_email = uid_manager.get('mail')
            if uid_manager_email:
                to_addrs.append(uid_manager_email)
        return to_addrs

    # eg, dn = uid=bill_lumbergh,cn=users,dc=initech,dc=com
    def get_metadata_from_dn(self, user_dn):
        if self.cache_engine:
            cache_result = self.caching.get(user_dn)
            if cache_result:
                cache_msg = 'Got ldap metadata from local cache for: %s' % user_dn
                self.log.debug(cache_msg)
                return cache_result
        ldap_filter = '(%s=*)' % self.uid_key
        ldap_results = self.search_ldap(user_dn, ldap_filter, attributes=self.attributes)
        if ldap_results:
            ldap_user_metadata = self.get_dict_from_ldap_object(self.connection.entries[0])
        else:
            self.caching.set(user_dn, {})
            return {}
        if self.cache_engine:
            self.log.debug('Writing user: %s metadata to cache engine.' % user_dn)
            self.caching.set(user_dn, ldap_user_metadata)
            self.caching.set(ldap_user_metadata[self.uid_key], ldap_user_metadata)
        return ldap_user_metadata

    def get_dict_from_ldap_object(self, ldap_user_object):
        ldap_user_metadata = {attr.key: attr.value for attr in ldap_user_object}
        ldap_user_metadata['dn'] = ldap_user_object.entry_dn

        email_key = ldap_user_metadata.get(self.email_key, None)
        uid_key = ldap_user_metadata.get(self.uid_key, None)

        if not email_key or not uid_key:
            return {}
        else:
            ldap_user_metadata['self.email_key'] = email_key.lower()
            ldap_user_metadata['self.uid_key'] = uid_key.lower()

        return ldap_user_metadata

    # eg, uid = bill_lumbergh
    def get_metadata_from_uid(self, uid):
        uid = uid.lower()
        if self.uid_regex:
            # for example if you set ldap_uid_regex in your mailer.yml to "^[0-9]{6}$" then it
            # would only query LDAP if your string length is 6 characters long and only digits.
            # re.search("^[0-9]{6}$", "123456")
            # Out[41]: <_sre.SRE_Match at 0x1109ab440>
            # re.search("^[0-9]{6}$", "1234567") returns None, or "12345a' also returns None
            if not re.search(self.uid_regex, uid):
                regex_msg = 'uid does not match regex: %s %s' % (self.uid_regex, uid)
                self.log.debug(regex_msg)
                return {}
        if self.cache_engine:
            cache_result = self.caching.get(uid)
            if cache_result or cache_result == {}:
                cache_msg = 'Got ldap metadata from local cache for: %s' % uid
                self.log.debug(cache_msg)
                return cache_result
        ldap_filter = '(%s=%s)' % (self.uid_key, uid)
        ldap_results = self.search_ldap(self.base_dn, ldap_filter, attributes=self.attributes)
        if ldap_results:
            ldap_user_metadata = self.get_dict_from_ldap_object(self.connection.entries[0])
            if self.cache_engine:
                self.log.debug('Writing user: %s metadata to cache engine.' % uid)
                if ldap_user_metadata.get('dn'):
                    self.caching.set(ldap_user_metadata['dn'], ldap_user_metadata)
                    self.caching.set(uid, ldap_user_metadata)
                else:
                    self.caching.set(uid, {})
        else:
            if self.cache_engine:
                self.caching.set(uid, {})
            return {}
        return ldap_user_metadata


# Use sqlite as a local cache for folks not running the mailer in lambda, avoids extra daemons
# as dependencies. This normalizes the methods to set/get functions, so you can interchangeable
# decide which caching system to use, a local file, or memcache, redis, etc
# If you don't want a redis dependency and aren't running the mailer in lambda this works well
class LocalSqlite(object):
    def __init__(self, local_filename, logger):
        self.log = logger
        self.sqlite = sqlite3.connect(local_filename)
        self.sqlite.execute('''CREATE TABLE IF NOT EXISTS ldap_cache(key text, value text)''')

    def get(self, key):
        sqlite_result = self.sqlite.execute("select * FROM ldap_cache WHERE key=?", (key,))
        result = sqlite_result.fetchall()
        if len(result) != 1:
            error_msg = 'Did not get 1 result from sqlite, something went wrong with key: %s' % key
            self.log.error(error_msg)
            return None
        return json.loads(result[0][1])

    def set(self, key, value):
        # note, the ? marks are required to ensure escaping into the database.
        self.sqlite.execute("INSERT INTO ldap_cache VALUES (?, ?)", (key, json.dumps(value)))
        self.sqlite.commit()


# redis can't write complex python objects like dictionaries as values (the way memcache can)
# so we turn our dict into a json string when setting, and json.loads when getting
class Redis(object):
    def __init__(self, redis_host=None, redis_port=6379, db=0):
        self.connection = redis.StrictRedis(host=redis_host, port=redis_port, db=db)

    def get(self, key):
        cache_value = self.connection.get(key)
        if cache_value:
            return json.loads(cache_value)

    def set(self, key, value):
        return self.connection.set(key, json.dumps(value))



"""
Allow local testing of mailer and templates by replaying an SQS message.

MAILER_FILE input is a file containing the exact base64-encoded, gzipped
data that's enqueued to SQS via :py:meth:`c7n.actions.Notify.send_sqs`.

Alternatively, with -p|--plain specified, the file will be assumed to be
JSON data that can be loaded directly.
"""
from __future__ import absolute_import, division, print_function, unicode_literals

import argparse
import boto3
import os
import logging
import zlib
import base64
import json

import jsonschema
from ruamel import yaml

from c7n_mailer.utils import setup_defaults
from c7n_mailer.cli import CONFIG_SCHEMA
from .email_delivery import EmailDelivery

logger = logging.getLogger(__name__)


class MailerTester(object):

    def __init__(self, msg_file, config, msg_plain=False, json_dump_file=None):
        if not os.path.exists(msg_file):
            raise RuntimeError("File does not exist: %s" % msg_file)
        logger.debug('Reading message from: %s', msg_file)
        with open(msg_file, 'r') as fh:
            raw = fh.read()
        logger.debug('Read %d byte message', len(raw))
        if msg_plain:
            raw = raw.strip()
        else:
            logger.debug('base64-decoding and zlib decompressing message')
            raw = zlib.decompress(base64.b64decode(raw))
            if json_dump_file is not None:
                with open(json_dump_file, 'wb') as fh:  # pragma: no cover
                    fh.write(raw)
        self.data = json.loads(raw)
        logger.debug('Loaded message JSON')
        self.config = config
        self.session = boto3.Session()

    def run(self, dry_run=False, print_only=False):
        emd = EmailDelivery(self.config, self.session, logger)
        addrs_to_msgs = emd.get_to_addrs_email_messages_map(self.data)
        logger.info('Would send email to: %s', addrs_to_msgs.keys())
        if print_only:
            mime = emd.get_mimetext_message(
                self.data, self.data['resources'], ['foo@example.com']
            )
            logger.info('Send mail with subject: "%s"', mime['Subject'])
            print(mime.get_payload(None, True))
            return
        if dry_run:
            for to_addrs, mimetext_msg in addrs_to_msgs.items():
                print('-> SEND MESSAGE TO: %s' % '; '.join(to_addrs))
                print(mimetext_msg.get_payload(None, True))
            return
        # else actually send the message...
        for to_addrs, mimetext_msg in addrs_to_msgs.items():
            logger.info('Actually sending mail to: %s', to_addrs)
            emd.send_c7n_email(self.data, list(to_addrs), mimetext_msg)


def setup_parser():
    parser = argparse.ArgumentParser('Test c7n-mailer templates and mail')
    parser.add_argument('-c', '--config', required=True)
    parser.add_argument('-d', '--dry-run', dest='dry_run', action='store_true',
                        default=False,
                        help='Log messages that would be sent, but do not send')
    parser.add_argument('-T', '--template-print', dest='print_only',
                        action='store_true', default=False,
                        help='Just print rendered templates')
    parser.add_argument('-t', '--templates', default=None, type=str,
                        help='message templates folder location')
    parser.add_argument('-p', '--plain', dest='plain', action='store_true',
                        default=False,
                        help='Expect MESSAGE_FILE to be a plain string, '
                             'rather than the base64-encoded, gzipped SQS '
                             'message format')
    parser.add_argument('-j', '--json-dump-file', dest='json_dump_file',
                        type=str, action='store', default=None,
                        help='If dump JSON of MESSAGE_FILE to this path; '
                             'useful to base64-decode and gunzip a message')
    parser.add_argument('MESSAGE_FILE', type=str,
                        help='Path to SQS message dump/content file')
    return parser


def session_factory(config):
    return boto3.Session(
        region_name=config['region'],
        profile_name=config.get('profile'))


def main():
    parser = setup_parser()
    options = parser.parse_args()

    module_dir = os.path.dirname(os.path.abspath(__file__))
    default_templates = [
        os.path.abspath(os.path.join(module_dir, 'msg-templates')),
        os.path.abspath(os.path.join(module_dir, '..', 'msg-templates')),
        os.path.abspath('.')
    ]
    templates = options.templates
    if templates:
        default_templates.append(
            os.path.abspath(os.path.expanduser(os.path.expandvars(templates)))
        )

    log_format = '%(asctime)s - %(name)s - %(levelname)s - %(message)s'
    logging.basicConfig(level=logging.DEBUG, format=log_format)
    logging.getLogger('botocore').setLevel(logging.WARNING)

    with open(options.config) as fh:
        config = yaml.load(fh.read(), Loader=yaml.SafeLoader)

    jsonschema.validate(config, CONFIG_SCHEMA)
    setup_defaults(config)
    config['templates_folders'] = default_templates

    tester = MailerTester(
        options.MESSAGE_FILE, config, msg_plain=options.plain,
        json_dump_file=options.json_dump_file
    )
    tester.run(options.dry_run, options.print_only)


if __name__ == '__main__':
    main()



# Copyright 2018 Capital One Services, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sendgrid
import six
from c7n_mailer.utils import (get_message_subject, get_rendered_jinja)
from c7n_mailer.utils_email import is_email
from python_http_client import exceptions
from sendgrid.helpers.mail import Mail, To, From


class SendGridDelivery(object):

    def __init__(self, config, logger):
        self.config = config
        self.logger = logger
        self.sendgrid_client = \
            sendgrid.SendGridAPIClient(self.config.get('sendgrid_api_key', ''))

    def get_to_addrs_sendgrid_messages_map(self, queue_message):
        # eg: { ('milton@initech.com', 'peter@initech.com'): [resource1, resource2, etc] }
        to_addrs_to_resources_map = self.get_email_to_addrs_to_resources_map(queue_message)

        to_addrs_to_content_map = {}
        for to_addrs, resources in six.iteritems(to_addrs_to_resources_map):
            to_addrs_to_content_map[to_addrs] = self.get_message_content(
                queue_message,
                resources,
                list(to_addrs)
            )
        # eg: { ('milton@initech.com', 'peter@initech.com'): message }
        return to_addrs_to_content_map

    # this function returns a dictionary with a tuple of emails as the key
    # and the list of resources as the value. This helps ensure minimal emails
    # are sent, while only ever sending emails to the respective parties.
    def get_email_to_addrs_to_resources_map(self, queue_message):
        email_to_addrs_to_resources_map = {}
        targets = queue_message['action']['to']

        for resource in queue_message['resources']:
            # this is the list of emails that will be sent for this resource
            resource_emails = []

            for target in targets:
                if target.startswith('tag:') and 'tags' in resource:
                    tag_name = target.split(':', 1)[1]
                    result = resource.get('tags', {}).get(tag_name, None)
                    if is_email(result):
                        resource_emails.append(result)
                elif is_email(target):
                    resource_emails.append(target)

            resource_emails = tuple(sorted(set(resource_emails)))

            if resource_emails:
                email_to_addrs_to_resources_map.setdefault(resource_emails, []).append(resource)

        if email_to_addrs_to_resources_map == {}:
            self.logger.debug('Found no email addresses, sending no emails.')
        # eg: { ('milton@initech.com', 'peter@initech.com'): [resource1, resource2, etc] }
        return email_to_addrs_to_resources_map

    def get_message_content(self, queue_message, resources, to_addrs):
        return get_rendered_jinja(
            to_addrs, queue_message, resources, self.logger,
            'template', 'default', self.config['templates_folders'])

    def sendgrid_handler(self, queue_message, to_addrs_to_email_messages_map):
        self.logger.info("Sending account:%s policy:%s %s:%s email:%s to %s" % (
            queue_message.get('account', ''),
            queue_message['policy']['name'],
            queue_message['policy']['resource'],
            str(len(queue_message['resources'])),
            queue_message['action'].get('template', 'default'),
            to_addrs_to_email_messages_map))

        from_address = self.config.get('from_address', '')
        subject = get_message_subject(queue_message)
        email_format = queue_message['action'].get('template_format', None)
        if not email_format:
            email_format = queue_message['action'].get(
                'template', 'default').endswith('html') and 'html' or 'plain'

        for email_to_addrs, email_content in six.iteritems(to_addrs_to_email_messages_map):
            for to_address in email_to_addrs:
                if email_format == "html":
                    mail = Mail(
                        from_email=From(from_address),
                        to_emails=To(to_address),
                        subject=subject,
                        html_content=email_content)
                else:
                    mail = Mail(
                        from_email=From(from_address),
                        to_emails=To(to_address),
                        subject=subject,
                        plain_text_content=email_content)
                try:
                    self.sendgrid_client.send(mail)
                except (exceptions.UnauthorizedError, exceptions.BadRequestsError) as e:
                    self.logger.warning(
                        "\n**Error \nPolicy:%s \nAccount:%s \nSending to:%s \n\nRequest body:"
                        "\n%s\n\nRequest headers:\n%s\n\n mailer.yml: %s" % (
                            queue_message['policy'],
                            queue_message.get('account', ''),
                            email_to_addrs,
                            e.body,
                            e.headers,
                            self.config
                        )
                    )
                    return False
        return True



from __future__ import absolute_import, division, print_function, unicode_literals

import argparse
import functools
import logging
from os import path

import boto3
import jsonschema
from c7n_mailer import deploy, utils
from c7n_mailer.azure.azure_queue_processor import MailerAzureQueueProcessor
from c7n_mailer.azure import deploy as azure_deploy
from c7n_mailer.sqs_queue_processor import MailerSqsQueueProcessor
from c7n_mailer.utils import get_provider, Providers
from ruamel import yaml

CONFIG_SCHEMA = {
    'type': 'object',
    'additionalProperties': False,
    'required': ['queue_url'],
    'properties': {
        'queue_url': {'type': 'string'},
        'from_address': {'type': 'string'},
        'contact_tags': {'type': 'array', 'items': {'type': 'string'}},
        'org_domain': {'type': 'string'},

        # Standard Lambda Function Config
        'region': {'type': 'string'},
        'role': {'type': 'string'},
        'runtime': {'type': 'string'},
        'memory': {'type': 'integer'},
        'timeout': {'type': 'integer'},
        'subnets': {'type': 'array', 'items': {'type': 'string'}},
        'security_groups': {'type': 'array', 'items': {'type': 'string'}},
        'dead_letter_config': {'type': 'object'},
        'lambda_name': {'type': 'string'},
        'lambda_description': {'type': 'string'},
        'lambda_tags': {'type': 'object'},
        'lambda_schedule': {'type': 'string'},

        # Azure Function Config
        'function_properties': {
            'type': 'object',
            'appInsights': {
                'type': 'object',
                'oneOf': [
                    {'type': 'string'},
                    {'type': 'object',
                        'properties': {
                            'name': 'string',
                            'location': 'string',
                            'resourceGroupName': 'string'}
                     }
                ]
            },
            'storageAccount': {
                'type': 'object',
                'oneOf': [
                    {'type': 'string'},
                    {'type': 'object',
                        'properties': {
                            'name': 'string',
                            'location': 'string',
                            'resourceGroupName': 'string'}
                     }
                ]
            },
            'servicePlan': {
                'type': 'object',
                'oneOf': [
                    {'type': 'string'},
                    {'type': 'object',
                        'properties': {
                            'name': 'string',
                            'location': 'string',
                            'resourceGroupName': 'string',
                            'skuTier': 'string',
                            'skuName': 'string'}
                     }
                ]
            },
        },
        'function_schedule': {'type': 'string'},
        'function_skuCode': {'type': 'string'},
        'function_sku': {'type': 'string'},

        # Mailer Infrastructure Config
        'cache_engine': {'type': 'string'},
        'smtp_server': {'type': 'string'},
        'smtp_port': {'type': 'integer'},
        'smtp_ssl': {'type': 'boolean'},
        'smtp_username': {'type': 'string'},
        'smtp_password': {'type': 'string'},
        'ldap_email_key': {'type': 'string'},
        'ldap_uid_tags': {'type': 'array', 'items': {'type': 'string'}},
        'debug': {'type': 'boolean'},
        'ldap_uid_regex': {'type': 'string'},
        'ldap_uri': {'type': 'string'},
        'ldap_bind_dn': {'type': 'string'},
        'ldap_bind_user': {'type': 'string'},
        'ldap_uid_attribute': {'type': 'string'},
        'ldap_manager_attribute': {'type': 'string'},
        'ldap_email_attribute': {'type': 'string'},
        'ldap_bind_password_in_kms': {'type': 'boolean'},
        'ldap_bind_password': {'type': 'string'},
        'cross_accounts': {'type': 'object'},
        'ses_region': {'type': 'string'},
        'redis_host': {'type': 'string'},
        'redis_port': {'type': 'integer'},
        'datadog_api_key': {'type': 'string'},              # TODO: encrypt with KMS?
        'datadog_application_key': {'type': 'string'},      # TODO: encrypt with KMS?
        'slack_token': {'type': 'string'},
        'slack_webhook': {'type': 'string'},
        'sendgrid_api_key': {'type': 'string'},
        'splunk_hec_url': {'type': 'string'},
        'splunk_hec_token': {'type': 'string'},
        'splunk_remove_paths': {
            'type': 'array',
            'items': {'type': 'string'}
        },
        'splunk_actions_list': {'type': 'boolean'},
        'splunk_max_attempts': {'type': 'integer'},
        'splunk_hec_max_length': {'type': 'integer'},

        # SDK Config
        'profile': {'type': 'string'},
        'http_proxy': {'type': 'string'},
        'https_proxy': {'type': 'string'},

        # Mapping account / emails
        'account_emails': {'type': 'object'}
    }
}


def session_factory(mailer_config):
    return boto3.Session(
        region_name=mailer_config['region'],
        profile_name=mailer_config.get('profile', None))


def get_logger(debug=False):
    log_format = '%(asctime)s - %(name)s - %(levelname)s - %(message)s'
    logging.basicConfig(level=logging.INFO, format=log_format)
    logging.getLogger('botocore').setLevel(logging.WARNING)
    if debug:
        logging.getLogger('botocore').setLevel(logging.DEBUG)
        debug_logger = logging.getLogger('custodian-mailer')
        debug_logger.setLevel(logging.DEBUG)
        return debug_logger
    else:
        return logging.getLogger('custodian-mailer')


def get_and_validate_mailer_config(args):
    with open(args.config) as fh:
        config = yaml.load(fh.read(), Loader=yaml.SafeLoader)
    jsonschema.validate(config, CONFIG_SCHEMA)
    utils.setup_defaults(config)
    return config


def get_c7n_mailer_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument('-c', '--config', required=True, help='mailer.yml config file')
    debug_help_msg = 'sets c7n_mailer logger to debug, for maximum output (the default is INFO)'
    parser.add_argument('--debug', action='store_true', help=debug_help_msg)
    max_num_processes_help_msg = 'will run the mailer in parallel, integer of max processes allowed'
    parser.add_argument('--max-num-processes', type=int, help=max_num_processes_help_msg)
    templates_folder_help_msg = 'message templates folder location'
    parser.add_argument('-t', '--templates', help=templates_folder_help_msg)
    group = parser.add_mutually_exclusive_group(required=True)
    update_lambda_help_msg = 'packages your c7n_mailer, uploads the zip to aws lambda as a function'
    group.add_argument('--update-lambda', action='store_true', help=update_lambda_help_msg)
    run_help_msg = 'run c7n-mailer locally, process sqs messages and send emails or sns messages'
    group.add_argument('--run', action='store_true', help=run_help_msg)
    return parser


def run_mailer_in_parallel(processor, max_num_processes):
    max_num_processes = int(max_num_processes)
    if max_num_processes < 1:
        raise Exception
    processor.max_num_processes = max_num_processes
    processor.run(parallel=True)


def main():
    parser = get_c7n_mailer_parser()
    args = parser.parse_args()
    mailer_config = get_and_validate_mailer_config(args)
    args_dict = vars(args)
    logger = get_logger(debug=args_dict.get('debug', False))

    module_dir = path.dirname(path.abspath(__file__))
    default_templates = [path.abspath(path.join(module_dir, 'msg-templates')),
                         path.abspath(path.join(module_dir, '..', 'msg-templates')),
                         path.abspath('.')]
    templates = args_dict.get('templates', None)
    if templates:
        default_templates.append(path.abspath(path.expanduser(path.expandvars(templates))))

    mailer_config['templates_folders'] = default_templates

    provider = get_provider(mailer_config)
    if args_dict.get('update_lambda'):
        if args_dict.get('debug'):
            print('\n** --debug is only supported with --run, not --update-lambda **\n')
            return
        if args_dict.get('max_num_processes'):
            print('\n** --max-num-processes is only supported '
                  'with --run, not --update-lambda **\n')
            return

        if provider == Providers.Azure:
            azure_deploy.provision(mailer_config)
        elif provider == Providers.AWS:
            deploy.provision(mailer_config, functools.partial(session_factory, mailer_config))

    if args_dict.get('run'):
        max_num_processes = args_dict.get('max_num_processes')

        # Select correct processor
        if provider == Providers.Azure:
            processor = MailerAzureQueueProcessor(mailer_config, logger)
        elif provider == Providers.AWS:
            aws_session = session_factory(mailer_config)
            processor = MailerSqsQueueProcessor(mailer_config, aws_session, logger)

        # Execute
        if max_num_processes:
            run_mailer_in_parallel(processor, max_num_processes)
        else:
            processor.run()


if __name__ == '__main__':
    main()



# Copyright 2017 Capital One Services, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from email.mime.text import MIMEText
from itertools import chain
import six

from .ldap_lookup import LdapLookup
from c7n_mailer.utils_email import is_email
from c7n_mailer.smtp_delivery import SmtpDelivery
from .utils import (
    get_message_subject, get_resource_tag_targets,
    get_rendered_jinja, kms_decrypt, get_aws_username_from_event)

# Those headers are defined as follows:
#  'X-Priority': 1 (Highest), 2 (High), 3 (Normal), 4 (Low), 5 (Lowest)
#              Non-standard, cf https://people.dsv.su.se/~jpalme/ietf/ietf-mail-attributes.html
#              Set by Thunderbird
#  'X-MSMail-Priority': High, Normal, Low
#              Cf Microsoft https://msdn.microsoft.com/en-us/library/gg671973(v=exchg.80).aspx
#              Note: May increase SPAM level on Spamassassin:
#                    https://wiki.apache.org/spamassassin/Rules/MISSING_MIMEOLE
#  'Priority': "normal" / "non-urgent" / "urgent"
#              Cf https://tools.ietf.org/html/rfc2156#section-5.3.6
#  'Importance': "low" / "normal" / "high"
#              Cf https://tools.ietf.org/html/rfc2156#section-5.3.4
PRIORITIES = {
    '1': {
        'X-Priority': '1 (Highest)',
        'X-MSMail-Priority': 'High',
        'Priority': 'urgent',
        'Importance': 'high',
    },
    '2': {
        'X-Priority': '2 (High)',
        'X-MSMail-Priority': 'High',
        'Priority': 'urgent',
        'Importance': 'high',
    },
    '3': {
        'X-Priority': '3 (Normal)',
        'X-MSMail-Priority': 'Normal',
        'Priority': 'normal',
        'Importance': 'normal',
    },
    '4': {
        'X-Priority': '4 (Low)',
        'X-MSMail-Priority': 'Low',
        'Priority': 'non-urgent',
        'Importance': 'low',
    },
    '5': {
        'X-Priority': '5 (Lowest)',
        'X-MSMail-Priority': 'Low',
        'Priority': 'non-urgent',
        'Importance': 'low',
    }
}


class EmailDelivery(object):

    def __init__(self, config, session, logger):
        self.config = config
        self.logger = logger
        self.session = session
        self.aws_ses = session.client('ses', region_name=config.get('ses_region'))
        self.ldap_lookup = self.get_ldap_connection()

    def get_ldap_connection(self):
        if self.config.get('ldap_uri'):
            self.config['ldap_bind_password'] = kms_decrypt(self.config, self.logger,
                                                            self.session, 'ldap_bind_password')
            return LdapLookup(self.config, self.logger)
        return None

    def priority_header_is_valid(self, priority_header):
        try:
            priority_header_int = int(priority_header)
        except ValueError:
            return False
        if priority_header_int and 0 < int(priority_header_int) < 6:
            return True
        else:
            self.logger.warning('mailer priority_header is not a valid string from 1 to 5')
            return False

    def get_valid_emails_from_list(self, targets):
        emails = []
        for target in targets:
            if is_email(target):
                emails.append(target)
        return emails

    def get_event_owner_email(self, targets, event):
        if 'event-owner' in targets:
            aws_username = get_aws_username_from_event(self.logger, event)
            if aws_username:
                # is using SSO, the target might already be an email
                if is_email(aws_username):
                    return [aws_username]
                # if the LDAP config is set, lookup in ldap
                elif self.config.get('ldap_uri', False):
                    return self.ldap_lookup.get_email_to_addrs_from_uid(aws_username)
                # the org_domain setting is configured, append the org_domain
                # to the username from AWS
                elif self.config.get('org_domain', False):
                    org_domain = self.config.get('org_domain', False)
                    self.logger.info('adding email %s to targets.', aws_username + '@' + org_domain)
                    return [aws_username + '@' + org_domain]
                else:
                    self.logger.warning('unable to lookup owner email. \
                            Please configure LDAP or org_domain')
            else:
                self.logger.info('no aws username in event')
        return []

    def get_ldap_emails_from_resource(self, sqs_message, resource):
        ldap_uid_tag_keys = self.config.get('ldap_uid_tags', [])
        ldap_uri = self.config.get('ldap_uri', False)
        if not ldap_uid_tag_keys or not ldap_uri:
            return []
        # this whole section grabs any ldap uids (including manager emails if option is on)
        # and gets the emails for them and returns an array with all the emails
        ldap_uid_tag_values = get_resource_tag_targets(resource, ldap_uid_tag_keys)
        email_manager = sqs_message['action'].get('email_ldap_username_manager', False)
        ldap_uid_emails = []
        # some types of resources, like iam-user have 'Username' in the resource, if the policy
        # opted in to resource_ldap_lookup_username: true, we'll do a lookup and send an email
        if sqs_message['action'].get('resource_ldap_lookup_username'):
            ldap_uid_emails = ldap_uid_emails + self.ldap_lookup.get_email_to_addrs_from_uid(
                resource.get('UserName'),
                manager=email_manager
            )
        for ldap_uid_tag_value in ldap_uid_tag_values:
            ldap_emails_set = self.ldap_lookup.get_email_to_addrs_from_uid(
                ldap_uid_tag_value,
                manager=email_manager
            )
            ldap_uid_emails = ldap_uid_emails + ldap_emails_set
        return ldap_uid_emails

    def get_resource_owner_emails_from_resource(self, sqs_message, resource):
        if 'resource-owner' not in sqs_message['action']['to']:
            return []
        resource_owner_tag_keys = self.config.get('contact_tags', [])
        resource_owner_tag_values = get_resource_tag_targets(resource, resource_owner_tag_keys)
        explicit_emails = self.get_valid_emails_from_list(resource_owner_tag_values)

        # resolve the contact info from ldap
        non_email_ids = list(set(resource_owner_tag_values).difference(explicit_emails))
        ldap_emails = list(chain.from_iterable([self.ldap_lookup.get_email_to_addrs_from_uid
                                              (uid) for uid in non_email_ids]))

        return list(chain(explicit_emails, ldap_emails))

    def get_account_emails(self, sqs_message):
        email_list = []

        if 'account-emails' not in sqs_message['action']['to']:
            return []

        account_id = sqs_message.get('account_id', None)
        self.logger.debug('get_account_emails for account_id: %s.', account_id)

        if account_id is not None:
            account_email_mapping = self.config.get('account_emails', {})
            self.logger.debug(
                'get_account_emails account_email_mapping: %s.', account_email_mapping)
            email_list = account_email_mapping.get(account_id, [])
            self.logger.debug('get_account_emails email_list: %s.', email_list)

        return self.get_valid_emails_from_list(email_list)

    # this function returns a dictionary with a tuple of emails as the key
    # and the list of resources as the value. This helps ensure minimal emails
    # are sent, while only ever sending emails to the respective parties.
    def get_email_to_addrs_to_resources_map(self, sqs_message):
        # policy_to_emails always get sent to any email msg that goes out
        # these were manually set by the policy writer in notify to section
        # or it's an email from an aws event username from an ldap_lookup
        email_to_addrs_to_resources_map = {}
        targets = sqs_message['action']['to'] + \
            (sqs_message['action']['cc'] if 'cc' in sqs_message['action'] else [])
        no_owner_targets = self.get_valid_emails_from_list(
            sqs_message['action'].get('owner_absent_contact', [])
        )
        # policy_to_emails includes event-owner if that's set in the policy notify to section
        policy_to_emails = self.get_valid_emails_from_list(targets)
        # if event-owner is set, and the aws_username has an ldap_lookup email
        # we add that email to the policy emails for these resource(s) on this sqs_message
        event_owner_email = self.get_event_owner_email(targets, sqs_message['event'])

        account_emails = self.get_account_emails(sqs_message)

        policy_to_emails = policy_to_emails + event_owner_email + account_emails
        for resource in sqs_message['resources']:
            # this is the list of emails that will be sent for this resource
            resource_emails = []
            # add in any ldap emails to resource_emails
            resource_emails = resource_emails + self.get_ldap_emails_from_resource(
                sqs_message,
                resource
            )
            resource_emails = resource_emails + policy_to_emails
            # add in any emails from resource-owners to resource_owners
            ro_emails = self.get_resource_owner_emails_from_resource(
                sqs_message,
                resource
            )

            resource_emails = resource_emails + ro_emails
            # if 'owner_absent_contact' was specified in the policy and no resource
            # owner emails were found, add those addresses
            if len(ro_emails) < 1 and len(no_owner_targets) > 0:
                resource_emails = resource_emails + no_owner_targets
            # we allow multiple emails from various places, we'll unique with set to not have any
            # duplicates, and we'll also sort it so we always have the same key for other resources
            # and finally we'll make it a tuple, since that is hashable and can be a key in a dict
            resource_emails = tuple(sorted(set(resource_emails)))
            # only if there are valid emails available, add it to the map
            if resource_emails:
                email_to_addrs_to_resources_map.setdefault(resource_emails, []).append(resource)
        if email_to_addrs_to_resources_map == {}:
            self.logger.debug('Found no email addresses, sending no emails.')
        # eg: { ('milton@initech.com', 'peter@initech.com'): [resource1, resource2, etc] }
        return email_to_addrs_to_resources_map

    def get_to_addrs_email_messages_map(self, sqs_message):
        to_addrs_to_resources_map = self.get_email_to_addrs_to_resources_map(sqs_message)
        to_addrs_to_mimetext_map = {}
        for to_addrs, resources in six.iteritems(to_addrs_to_resources_map):
            to_addrs_to_mimetext_map[to_addrs] = self.get_mimetext_message(
                sqs_message,
                resources,
                list(to_addrs)
            )
        # eg: { ('milton@initech.com', 'peter@initech.com'): mimetext_message }
        return to_addrs_to_mimetext_map

    def set_mimetext_headers(self, message, subject, from_addr, to_addrs, cc_addrs, priority):
        """Sets headers on Mimetext message"""

        message['Subject'] = subject
        message['From'] = from_addr
        message['To'] = ', '.join(to_addrs)
        if cc_addrs:
            message['Cc'] = ', '.join(cc_addrs)

        if priority and self.priority_header_is_valid(priority):
            priority = PRIORITIES[str(priority)].copy()
            for key in priority:
                message[key] = priority[key]

        return message

    def get_mimetext_message(self, sqs_message, resources, to_addrs):
        body = get_rendered_jinja(
            to_addrs, sqs_message, resources, self.logger,
            'template', 'default', self.config['templates_folders'])

        if not body:
            return None

        email_format = sqs_message['action'].get('template_format', None)
        if not email_format:
            email_format = sqs_message['action'].get(
                'template', 'default').endswith('html') and 'html' or 'plain'

        message = self.set_mimetext_headers(
            message=MIMEText(body, email_format, 'utf-8'),
            subject=get_message_subject(sqs_message),
            from_addr=sqs_message['action'].get('from', self.config['from_address']),
            to_addrs=to_addrs,
            cc_addrs=sqs_message['action'].get('cc', []),
            priority=sqs_message['action'].get('priority_header', None),
        )

        return message

    def send_c7n_email(self, sqs_message, email_to_addrs, mimetext_msg):
        try:
            # if smtp_server is set in mailer.yml, send through smtp
            if 'smtp_server' in self.config:
                smtp_delivery = SmtpDelivery(config=self.config,
                                             session=self.session,
                                             logger=self.logger)
                smtp_delivery.send_message(message=mimetext_msg, to_addrs=email_to_addrs)
            # if smtp_server isn't set in mailer.yml, use aws ses normally.
            else:
                self.aws_ses.send_raw_email(RawMessage={'Data': mimetext_msg.as_string()})
        except Exception as error:
            self.logger.warning(
                "Error policy:%s account:%s sending to:%s \n\n error: %s\n\n mailer.yml: %s" % (
                    sqs_message['policy'],
                    sqs_message.get('account', ''),
                    email_to_addrs,
                    error,
                    self.config
                )
            )
        self.logger.info("Sending account:%s policy:%s %s:%s email:%s to %s" % (
            sqs_message.get('account', ''),
            sqs_message['policy']['name'],
            sqs_message['policy']['resource'],
            str(len(sqs_message['resources'])),
            sqs_message['action'].get('template', 'default'),
            email_to_addrs))



# Copyright 2017 Capital One Services, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time

from datadog import api
from datadog import initialize
from six.moves.urllib.parse import urlparse, parse_qsl


class DataDogDelivery(object):
    DATADOG_API_KEY = 'datadog_api_key'
    DATADOG_APPLICATION_KEY = 'datadog_application_key'

    def __init__(self, config, session, logger):
        self.config = config
        self.logger = logger
        self.session = session
        self.datadog_api_key = self.config.get(self.DATADOG_API_KEY, None)
        self.datadog_application_key = self.config.get(self.DATADOG_APPLICATION_KEY, None)

        # Initialize datadog
        if self.datadog_api_key and self.datadog_application_key:
            options = {
                'api_key': self.datadog_api_key,
                'app_key': self.datadog_application_key,

            }
            initialize(**options)

    def get_datadog_message_packages(self, sqs_message):
        date_time = time.time()
        datadog_rendered_messages = []

        metric_config_map = self._get_metrics_config_to_resources_map(sqs_message)
        if not metric_config_map:
            return datadog_rendered_messages

        if sqs_message and sqs_message.get('resources', False):
            for resource in sqs_message['resources']:
                tags = [
                    'event:{}'.format(sqs_message['event']),
                    'account_id:{}'.format(sqs_message['account_id']),
                    'account:{}'.format(sqs_message['account']),
                    'region:{}'.format(sqs_message['region'])
                ]

                tags.extend(['{key}:{value}'.format(
                    key=key, value=resource[key]) for key in resource.keys()
                    if key != 'Tags'])
                if resource.get('Tags', False):
                    tags.extend(['{key}:{value}'.format(
                        key=tag['Key'], value=tag['Value']) for tag in resource['Tags']])

                for metric_config in metric_config_map:
                    datadog_rendered_messages.append({
                        "metric": metric_config['metric_name'],
                        "points": (date_time, self._get_metric_value(
                            metric_config=metric_config, tags=tags)),
                        "tags": tags
                    })

        # eg: [{'metric': 'metric_name', 'points': (date_time, value),
        # 'tags': ['tag1':'value', 'tag2':'value']}, ...]
        return datadog_rendered_messages

    def deliver_datadog_messages(self, datadog_message_packages, sqs_message):
        if len(datadog_message_packages) > 0:
            self.logger.info(
                "Sending account:{account} policy:{policy} {resource}:{quantity} to DataDog".
                format(account=sqs_message.get('account', ''),
                       policy=sqs_message['policy']['name'],
                       resource=sqs_message['policy']['resource'],
                       quantity=len(sqs_message['resources'])))

            api.Metric.send(datadog_message_packages)

    @staticmethod
    def _get_metric_value(metric_config, tags):
        metric_value = 1
        metric_value_tag = metric_config.get('metric_value_tag', 'default')
        if metric_value_tag != 'default':
            for tag in tags:
                if metric_value_tag in tag:
                    metric_value = float(tag[tag.find(":") + 1:])

        return metric_value

    @staticmethod
    def _get_metrics_config_to_resources_map(sqs_message):
        metric_config_map = []
        if sqs_message and sqs_message.get(
                'action', False) and sqs_message['action'].get('to', False):
            for to in sqs_message['action']['to']:
                if to.startswith('datadog://'):
                    parsed = urlparse(to)
                    metric_config_map.append(dict(parse_qsl(parsed.query)))
        return metric_config_map



import logging
from email.utils import parseaddr

logger = logging.getLogger('c7n_mailer.utils.email')


def is_email(target):
    if target is None:
        return False
    if target.startswith('slack://'):
        logger.debug("Slack payload, not an email.")
        return False
    if parseaddr(target)[1] and '@' in target and '.' in target:
        return True
    else:
        return False



# Copyright 2018 Capital One Services, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Azure Queue Message Processing
==============================

"""
import base64
import json
import traceback
import zlib

import six
from c7n_mailer.smtp_delivery import SmtpDelivery
from c7n_mailer.azure.sendgrid_delivery import SendGridDelivery

try:
    from c7n_azure.storage_utils import StorageUtilities
    from c7n_azure.session import Session
except ImportError:
    StorageUtilities = None
    Session = None
    pass


class MailerAzureQueueProcessor(object):

    def __init__(self, config, logger, session=None, max_num_processes=16):
        if StorageUtilities is None:
            raise Exception("Using Azure queue requires package c7n_azure to be installed.")

        self.max_num_processes = max_num_processes
        self.config = config
        self.logger = logger
        self.receive_queue = self.config['queue_url']
        self.batch_size = 16
        self.max_message_retry = 3
        self.session = session or Session()

    def run(self, parallel=False):
        if parallel:
            self.logger.info("Parallel processing with Azure Queue is not yet implemented.")

        self.logger.info("Downloading messages from the Azure Storage queue.")
        queue_settings = StorageUtilities.get_queue_client_by_uri(self.receive_queue, self.session)
        queue_messages = StorageUtilities.get_queue_messages(
            *queue_settings, num_messages=self.batch_size)

        while len(queue_messages) > 0:
            for queue_message in queue_messages:
                self.logger.debug("Message id: %s received" % queue_message.id)

                if (self.process_azure_queue_message(queue_message) or
                        queue_message.dequeue_count > self.max_message_retry):
                    # If message handled successfully or max retry hit, delete
                    StorageUtilities.delete_queue_message(*queue_settings, message=queue_message)

            queue_messages = StorageUtilities.get_queue_messages(
                *queue_settings, num_messages=self.batch_size)

        self.logger.info('No messages left on the azure storage queue, exiting c7n_mailer.')

    def process_azure_queue_message(self, encoded_azure_queue_message):
        queue_message = json.loads(
            zlib.decompress(base64.b64decode(encoded_azure_queue_message.content)))

        self.logger.debug("Got account:%s message:%s %s:%d policy:%s recipients:%s" % (
            queue_message.get('account', 'na'),
            encoded_azure_queue_message.id,
            queue_message['policy']['resource'],
            len(queue_message['resources']),
            queue_message['policy']['name'],
            ', '.join(queue_message['action'].get('to'))))

        if any(e.startswith('slack') or e.startswith('https://hooks.slack.com/')
                for e in queue_message.get('action', ()).get('to')):
            from c7n_mailer.slack_delivery import SlackDelivery
            slack_delivery = SlackDelivery(self.config,
                                           self.logger,
                                           SendGridDelivery(self.config, self.logger))
            slack_messages = slack_delivery.get_to_addrs_slack_messages_map(queue_message)
            try:
                slack_delivery.slack_handler(queue_message, slack_messages)
            except Exception:
                traceback.print_exc()
                pass

        # this section gets the map of metrics to send to datadog and delivers it
        if any(e.startswith('datadog') for e in queue_message.get('action', ()).get('to')):
            from c7n_mailer.datadog_delivery import DataDogDelivery
            datadog_delivery = DataDogDelivery(self.config, self.session, self.logger)
            datadog_message_packages = datadog_delivery.get_datadog_message_packages(queue_message)

            try:
                datadog_delivery.deliver_datadog_messages(datadog_message_packages, queue_message)
            except Exception:
                traceback.print_exc()
                pass

        # this section sends a notification to the resource owner via SendGrid
        try:
            sendgrid_delivery = SendGridDelivery(self.config, self.logger)
            email_messages = sendgrid_delivery.get_to_addrs_sendgrid_messages_map(queue_message)

            if 'smtp_server' in self.config:
                smtp_delivery = SmtpDelivery(config=self.config,
                                             session=self.session,
                                             logger=self.logger)
                for to_addrs, message in six.iteritems(email_messages):
                    smtp_delivery.send_message(message=message, to_addrs=list(to_addrs))
            else:
                return sendgrid_delivery.sendgrid_handler(queue_message, email_messages)
        except Exception:
            traceback.print_exc()



# Copyright 2016-2017 Capital One Services, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Lambda entry point
"""
from __future__ import absolute_import, division, print_function, unicode_literals

from c7n_azure.session import Session
from c7n_azure.constants import RESOURCE_STORAGE
from c7n_mailer.azure.azure_queue_processor import MailerAzureQueueProcessor


def start_c7n_mailer(logger, config, auth_file):
    try:
        logger.info('c7n_mailer starting...')
        session = Session(authorization_file=auth_file, resource=RESOURCE_STORAGE)
        mailer_azure_queue_processor = MailerAzureQueueProcessor(config, logger, session=session)
        mailer_azure_queue_processor.run()
    except Exception as e:
        logger.exception("Error starting mailer MailerAzureQueueProcessor(). \n Error: %s \n" % (e))



# Copyright 2016-2017 Capital One Services, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import absolute_import, division, print_function, unicode_literals

import copy
import json
import os

from c7n.mu import (
    CloudWatchEventSource,
    LambdaFunction,
    LambdaManager,
    PythonPackageArchive)


entry_source = """\
import logging

from c7n_mailer import handle

logger = logging.getLogger('custodian.mailer')
log_format = '%(asctime)s - %(name)s - %(levelname)s - %(message)s'
logging.basicConfig(level=logging.INFO, format=log_format)
logging.getLogger('botocore').setLevel(logging.WARNING)

def dispatch(event, context):
    return handle.start_c7n_mailer(logger)
"""


def get_archive(config):
    archive = PythonPackageArchive(modules=[
        'c7n_mailer',
        # core deps
        'jinja2', 'markupsafe', 'ruamel', 'ldap3', 'pyasn1', 'redis',
        # for other dependencies
        'pkg_resources',
        # transport datadog - recursive deps
        'datadog', 'simplejson', 'decorator',
        # requests (recursive deps), needed by datadog, slackclient, splunk
        'requests', 'urllib3', 'idna', 'chardet', 'certifi',
        # used by splunk; also dependencies of c7n itself
        'jsonpointer', 'jsonpatch'])

    for d in set(config['templates_folders']):
        if not os.path.exists(d):
            continue
        for t in [f for f in os.listdir(d) if os.path.splitext(f)[1] == '.j2']:
            with open(os.path.join(d, t)) as fh:
                archive.add_contents('msg-templates/%s' % t, fh.read())

    function_config = copy.deepcopy(config)
    function_config['templates_folders'] = ['msg-templates/']
    archive.add_contents('config.json', json.dumps(function_config))
    archive.add_contents('periodic.py', entry_source)

    archive.close()
    return archive


def provision(config, session_factory):
    func_config = dict(
        name=config.get('lambda_name', 'cloud-custodian-mailer'),
        description=config.get('lambda_description', 'Cloud Custodian Mailer'),
        tags=config.get('lambda_tags', {}),
        handler='periodic.dispatch',
        runtime=config['runtime'],
        memory_size=config['memory'],
        timeout=config['timeout'],
        role=config['role'],
        subnets=config['subnets'],
        security_groups=config['security_groups'],
        dead_letter_config=config.get('dead_letter_config', {}),
        events=[
            CloudWatchEventSource(
                {'type': 'periodic',
                 'schedule': config.get('lambda_schedule', 'rate(5 minutes)')},
                session_factory)
        ])

    archive = get_archive(config)
    func = LambdaFunction(func_config, archive)
    manager = LambdaManager(session_factory)
    manager.publish(func)



# Copyright 2019 Microsoft Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import smtplib
import c7n_mailer.utils as utils


class SmtpDelivery(object):

    def __init__(self, config, session, logger):
        smtp_server = config['smtp_server']
        smtp_port = int(config.get('smtp_port', 25))
        smtp_ssl = bool(config.get('smtp_ssl', True))
        smtp_username = config.get('smtp_username')
        smtp_password = utils.decrypt(config, logger, session, 'smtp_password')

        smtp_connection = smtplib.SMTP(smtp_server, smtp_port)
        if smtp_ssl:
            smtp_connection.starttls()
            smtp_connection.ehlo()

        if smtp_username or smtp_password:
            smtp_connection.login(smtp_username, smtp_password)

        self._smtp_connection = smtp_connection

    def __del__(self):
        self._smtp_connection.quit()

    def send_message(self, message, to_addrs):
        self._smtp_connection.sendmail(message['From'], to_addrs, message.as_string())



# Copyright 2016-2017 Capital One Services, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Lambda entry point
"""
from __future__ import absolute_import, division, print_function, unicode_literals

import boto3
import json
import os

from .sqs_queue_processor import MailerSqsQueueProcessor


def config_setup(config=None):
    task_dir = os.environ.get('LAMBDA_TASK_ROOT')
    os.environ['PYTHONPATH'] = "%s:%s" % (task_dir, os.environ.get('PYTHONPATH', ''))
    if not config:
        with open(os.path.join(task_dir, 'config.json')) as fh:
            config = json.load(fh)
    if 'http_proxy' in config:
        os.environ['http_proxy'] = config['http_proxy']
    if 'https_proxy' in config:
        os.environ['https_proxy'] = config['https_proxy']
    return config


def start_c7n_mailer(logger, config=None, parallel=False):
    try:
        session = boto3.Session()
        if not config:
            config = config_setup()
        logger.info('c7n_mailer starting...')
        mailer_sqs_queue_processor = MailerSqsQueueProcessor(config, session, logger)
        mailer_sqs_queue_processor.run(parallel)
    except Exception as e:
        logger.exception("Error starting mailer MailerSqsQueueProcessor(). \n Error: %s \n" % (e))



# Copyright 2017 Capital One Services, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import six
from boto3 import Session

from .utils import (
    get_message_subject, get_resource_tag_targets, get_rendered_jinja)


class SnsDelivery(object):

    def __init__(self, config, session, logger):
        self.config = config
        self.logger = logger
        self.aws_sts = session.client('sts')
        self.sns_cache = {}

    def deliver_sns_messages(self, packaged_sns_messages, sqs_message):
        for packaged_sns_message in packaged_sns_messages:
            topic = packaged_sns_message['topic']
            subject = packaged_sns_message['subject']
            sns_message = packaged_sns_message['sns_message']
            self.deliver_sns_message(topic, subject, sns_message, sqs_message)

    def get_valid_sns_from_list(self, possible_sns_values):
        sns_addresses = []
        for target in possible_sns_values:
            if self.target_is_sns(target):
                sns_addresses.append(target)
        return sns_addresses

    def get_sns_message_package(self, sqs_message, policy_sns_address, subject, resources):
        rendered_jinja_body = get_rendered_jinja(
            policy_sns_address,
            sqs_message,
            resources,
            self.logger,
            'template',
            'default',
            self.config['templates_folders']
        )
        return {
            'topic': policy_sns_address,
            'subject': subject,
            'sns_message': rendered_jinja_body
        }

    def get_sns_message_packages(self, sqs_message):
        sns_to_resources_map = self.get_sns_addrs_to_resources_map(sqs_message)
        subject = get_message_subject(sqs_message)
        sns_addrs_to_rendered_jinja_messages = []
        # take the map with lists of resources, and jinja render them and add them
        # to sns_addrs_to_rendered_jinja_messages as an sns_message package
        for sns_topic, resources in six.iteritems(sns_to_resources_map):
            sns_addrs_to_rendered_jinja_messages.append(
                self.get_sns_message_package(
                    sqs_message,
                    sns_topic,
                    subject,
                    resources
                )
            )

        if sns_addrs_to_rendered_jinja_messages == []:
            self.logger.debug('Found no sns addresses, delivering no messages.')
        return sns_addrs_to_rendered_jinja_messages

    def get_sns_addrs_to_resources_map(self, sqs_message):
        policy_to_sns_addresses = self.get_valid_sns_from_list(sqs_message['action'].get('to', []))
        sns_addrs_to_resources_map = {}
        # go over all sns_addresses from the to field
        for policy_sns_address in policy_to_sns_addresses:
            sns_addrs_to_resources_map[policy_sns_address] = sqs_message['resources']
        # get sns topics / messages inside resource-owners tags
        for resource in sqs_message['resources']:
            resource_owner_tag_keys = self.config.get('contact_tags', [])
            possible_sns_tag_values = get_resource_tag_targets(
                resource,
                resource_owner_tag_keys
            )
            sns_tag_values = self.get_valid_sns_from_list(possible_sns_tag_values)
            # for each resource, get any valid sns topics, and add them to the map
            for sns_tag_value in sns_tag_values:
                # skip sns topics in tags if they're already in the to field
                if sns_tag_value in policy_to_sns_addresses:
                    continue
                sns_addrs_to_resources_map.setdefault(sns_tag_value, []).append(resource)
        return sns_addrs_to_resources_map

    def target_is_sns(self, target):
        if target.startswith('arn:aws:sns'):
            return True
        return False

    def deliver_sns_message(self, topic, subject, rendered_jinja_body, sqs_message):
        # Max length of subject in sns is 100 chars
        if len(subject) > 100:
            subject = subject[:97] + '..'
        try:
            account = topic.split(':')[4]
            if account in self.sns_cache:
                sns = self.sns_cache[account]
            else:
                # if cross_accounts isn't set, we'll try using the current credentials
                if account not in self.config.get('cross_accounts', []):
                    session = Session()
                else:
                    creds = self.aws_sts.assume_role(
                        RoleArn=self.config['cross_accounts'][account],
                        RoleSessionName="CustodianNotification")['Credentials']
                    session = Session(
                        aws_access_key_id=creds['AccessKeyId'],
                        aws_secret_access_key=creds['SecretAccessKey'],
                        aws_session_token=creds['SessionToken'])
                self.sns_cache[account] = sns = session.client('sns')

            self.logger.info("Sending account:%s policy:%s sns:%s to %s" % (
                sqs_message.get('account', ''),
                sqs_message['policy']['name'],
                sqs_message['action'].get('template', 'default'),
                topic))
            sns.publish(TopicArn=topic, Subject=subject, Message=rendered_jinja_body)
        except Exception as e:
            self.logger.warning(
                "Error policy:%s account:%s sending sns to %s \n %s" % (
                    sqs_message['policy'], sqs_message.get('account', 'na'), topic, e))



# Copyright 2018 Capital One Services, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import sys
import json
from os.path import dirname, join

# The working path for the Azure Function doesn't include this file's folder
sys.path.append(dirname(dirname(__file__)))

from c7n_mailer.azure import handle

def main(input):
    logger = logging.getLogger('custodian.mailer')
    config_file = join(dirname(__file__), 'config.json')
    with open(config_file) as fh:
        config = json.load(fh)
    return handle.start_c7n_mailer(logger, config, join(dirname(__file__), 'auth.json'))

# flake8: noqa


# Copyright 2015-2017 Capital One Services, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import absolute_import, division, print_function, unicode_literals

from io import open
from os import path
from setuptools import setup, find_packages

# *Any* updates here should also go into c7n_mailer/deploy.py for lambda packaging.
requires = [
    "Jinja2",
    "boto3",
    "jsonschema",
    "simplejson",  # datadog packaging dep
    "python-dateutil>=2.8",
    "ruamel.yaml>=0.15.93",
    "datadog",
    "sendgrid==6.0.5",
    "ldap3",
    "c7n",
    "redis"]


try:
    from concurrent import futures  # noqa F401
except ImportError:
    # The backport has SyntaxErrors under py36, so avoid installing it.
    # https://github.com/agronholm/pythonfutures/issues/41
    requires += ['futures']

# read the contents of your README file
this_directory = path.abspath(path.dirname(__file__))
readme = path.join(this_directory, 'README.md')
long_description = ''
if path.exists(readme):
    with open(readme, encoding='utf-8') as f:
        long_description = f.read()

setup(
    name="c7n_mailer",
    version='0.5.2',
    description="Cloud Custodian - Reference Mailer",
    long_description=long_description,
    long_description_content_type='text/markdown',
    classifiers=[
        "Topic :: System :: Systems Administration",
        "Topic :: System :: Distributed Computing"
    ],
    url="https://github.com/cloud-custodian/cloud-custodian",
    license="Apache-2.0",
    packages=find_packages(),
    entry_points={
        'console_scripts': [
            'c7n-mailer = c7n_mailer.cli:main',
            'c7n-mailer-replay = c7n_mailer.replay:main'
        ]
    },
    install_requires=requires,
    package_data={str(''): [str('msg-templates/*.j2')]},
)



# Copyright 2018 Capital One Services, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import time

import requests
import six
from c7n_mailer.ldap_lookup import Redis
from c7n_mailer.utils import get_rendered_jinja
from c7n_mailer.utils_email import is_email


class SlackDelivery(object):

    def __init__(self, config, logger, email_handler):
        self.caching = self.cache_factory(config, config.get('cache_engine', None))
        self.config = config
        self.logger = logger
        self.email_handler = email_handler

    def cache_factory(self, config, type):
        if type == 'redis':
            return Redis(redis_host=config.get('redis_host'),
                         redis_port=int(config.get('redis_port', 6379)), db=0)
        else:
            return None

    def get_to_addrs_slack_messages_map(self, sqs_message):
        resource_list = []
        for resource in sqs_message['resources']:
            resource_list.append(resource)

        slack_messages = {}

        # Check for Slack targets in 'to' action and render appropriate template.
        for target in sqs_message.get('action', ()).get('to'):
            if target == 'slack://owners':
                to_addrs_to_resources_map = \
                    self.email_handler.get_email_to_addrs_to_resources_map(sqs_message)
                for to_addrs, resources in six.iteritems(to_addrs_to_resources_map):

                    resolved_addrs = self.retrieve_user_im(list(to_addrs))

                    if not resolved_addrs:
                        continue

                    for address, slack_target in resolved_addrs.items():
                        slack_messages[address] = get_rendered_jinja(
                            slack_target, sqs_message, resources,
                            self.logger, 'slack_template', 'slack_default',
                            self.config['templates_folders'])
                self.logger.debug(
                    "Generating messages for recipient list produced by resource owner resolution.")
            elif target.startswith('https://hooks.slack.com/'):
                slack_messages[target] = get_rendered_jinja(
                    target, sqs_message,
                    resource_list,
                    self.logger, 'slack_template', 'slack_default',
                    self.config['templates_folders'])
            elif target.startswith('slack://webhook/#') and self.config.get('slack_webhook'):
                webhook_target = self.config.get('slack_webhook')
                slack_messages[webhook_target] = get_rendered_jinja(
                    target.split('slack://webhook/#', 1)[1], sqs_message,
                    resource_list,
                    self.logger, 'slack_template', 'slack_default',
                    self.config['templates_folders'])
                self.logger.debug(
                    "Generating message for webhook %s." % self.config.get('slack_webhook'))
            elif target.startswith('slack://') and is_email(target.split('slack://', 1)[1]):
                resolved_addrs = self.retrieve_user_im([target.split('slack://', 1)[1]])
                for address, slack_target in resolved_addrs.items():
                    slack_messages[address] = get_rendered_jinja(
                        slack_target, sqs_message, resource_list,
                        self.logger, 'slack_template', 'slack_default',
                        self.config['templates_folders'])
            elif target.startswith('slack://#'):
                resolved_addrs = target.split('slack://#', 1)[1]
                slack_messages[resolved_addrs] = get_rendered_jinja(
                    resolved_addrs, sqs_message,
                    resource_list,
                    self.logger, 'slack_template', 'slack_default',
                    self.config['templates_folders'])
            elif target.startswith('slack://tag/') and 'tags' in resource:
                tag_name = target.split('tag/', 1)[1]
                result = resource.get('tags', {}).get(tag_name, None)
                resolved_addrs = result
                slack_messages[resolved_addrs] = get_rendered_jinja(
                    resolved_addrs, sqs_message,
                    resource_list,
                    self.logger, 'slack_template', 'slack_default',
                    self.config['templates_folders'])
                self.logger.debug("Generating message for specified Slack channel.")
        return slack_messages

    def slack_handler(self, sqs_message, slack_messages):
        for key, payload in slack_messages.items():
            self.logger.info("Sending account:%s policy:%s %s:%s slack:%s to %s" % (
                sqs_message.get('account', ''),
                sqs_message['policy']['name'],
                sqs_message['policy']['resource'],
                str(len(sqs_message['resources'])),
                sqs_message['action'].get('slack_template', 'slack_default'),
                key)
            )

            self.send_slack_msg(key, payload)

    def retrieve_user_im(self, email_addresses):
        list = {}

        if not self.config['slack_token']:
            self.logger.info("No Slack token found.")

        for address in email_addresses:
            if self.caching and self.caching.get(address):
                self.logger.debug('Got Slack metadata from cache for: %s' % address)
                list[address] = self.caching.get(address)
                continue

            response = requests.post(
                url='https://slack.com/api/users.lookupByEmail',
                data={'email': address},
                headers={'Content-Type': 'application/x-www-form-urlencoded',
                         'Authorization': 'Bearer %s' % self.config.get('slack_token')}).json()

            if not response["ok"]:
                if "headers" in response.keys() and "Retry-After" in response["headers"]:
                    self.logger.info(
                        "Slack API rate limiting. Waiting %d seconds",
                        int(response.headers['retry-after']))
                    time.sleep(int(response.headers['Retry-After']))
                    continue
                elif response["error"] == "invalid_auth":
                    raise Exception("Invalid Slack token.")
                elif response["error"] == "users_not_found":
                    self.logger.info("Slack user ID not found.")
                    if self.caching:
                        self.caching.set(address, {})
                    continue
                else:
                    self.logger.warning("Slack Response: {}".format(response))
            else:
                slack_user_id = response['user']['id']
                if 'enterprise_user' in response['user'].keys():
                    slack_user_id = response['user']['enterprise_user']['id']
                self.logger.debug(
                    "Slack account %s found for user %s", slack_user_id, address)
                if self.caching:
                    self.logger.debug('Writing user: %s metadata to cache.', address)
                    self.caching.set(address, slack_user_id)

                list[address] = slack_user_id

        return list

    def send_slack_msg(self, key, message_payload):

        if key.startswith('https://hooks.slack.com/'):
            response = requests.post(
                url=key,
                data=message_payload,
                headers={'Content-Type': 'application/json'})
        else:
            response = requests.post(
                url='https://slack.com/api/chat.postMessage',
                data=message_payload,
                headers={'Content-Type': 'application/json;charset=utf-8',
                         'Authorization': 'Bearer %s' % self.config.get('slack_token')})

        if response.status_code == 429 and "Retry-After" in response.headers:
            self.logger.info(
                "Slack API rate limiting. Waiting %d seconds",
                int(response.headers['retry-after']))
            time.sleep(int(response.headers['Retry-After']))
            return
        elif response.status_code != 200:  # pragma: no cover
            self.logger.info(
                "Error in sending Slack message status:%s response: %s",
                response.status_code, response.text())
            return



# Copyright 2015-2017 Capital One Services, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import absolute_import, division, print_function, unicode_literals

import base64
from datetime import datetime, timedelta
import functools
import json
import os
import time

import jinja2
import jmespath
from botocore.exceptions import ClientError
from dateutil import parser
from dateutil.tz import gettz, tzutc
from ruamel import yaml


class Providers(object):
    AWS = 0
    Azure = 1


def get_jinja_env(template_folders):
    env = jinja2.Environment(trim_blocks=True, autoescape=False)
    env.filters['yaml_safe'] = functools.partial(yaml.safe_dump, default_flow_style=False)
    env.filters['date_time_format'] = date_time_format
    env.filters['get_date_time_delta'] = get_date_time_delta
    env.filters['get_date_age'] = get_date_age
    env.globals['format_resource'] = resource_format
    env.globals['format_struct'] = format_struct
    env.globals['resource_tag'] = get_resource_tag_value
    env.globals['get_resource_tag_value'] = get_resource_tag_value
    env.globals['search'] = jmespath.search
    env.loader = jinja2.FileSystemLoader(template_folders)
    return env


def get_rendered_jinja(
        target, sqs_message, resources, logger,
        specified_template, default_template, template_folders):
    env = get_jinja_env(template_folders)
    mail_template = sqs_message['action'].get(specified_template, default_template)
    if not os.path.isabs(mail_template):
        mail_template = '%s.j2' % mail_template
    try:
        template = env.get_template(mail_template)
    except Exception as error_msg:
        logger.error("Invalid template reference %s\n%s" % (mail_template, error_msg))
        return

    # recast seconds since epoch as utc iso datestring, template
    # authors can use date_time_format helper func to convert local
    # tz. if no execution start time was passed use current time.
    execution_start = datetime.utcfromtimestamp(
        sqs_message.get(
            'execution_start',
            time.mktime(
                datetime.utcnow().timetuple())
        )).isoformat()

    rendered_jinja = template.render(
        recipient=target,
        resources=resources,
        account=sqs_message.get('account', ''),
        account_id=sqs_message.get('account_id', ''),
        event=sqs_message.get('event', None),
        action=sqs_message['action'],
        policy=sqs_message['policy'],
        execution_start=execution_start,
        region=sqs_message.get('region', ''))
    return rendered_jinja


# eg, target_tag_keys could be resource-owners ['Owners', 'SupportTeam']
# and this function would go through the resource and look for any tag keys
# that match Owners or SupportTeam, and return those values as targets
def get_resource_tag_targets(resource, target_tag_keys):
    if 'Tags' not in resource:
        return []
    tags = {tag['Key']: tag['Value'] for tag in resource['Tags']}
    targets = []
    for target_tag_key in target_tag_keys:
        if target_tag_key in tags:
            targets.append(tags[target_tag_key])
    return targets


def get_message_subject(sqs_message):
    default_subject = 'Custodian notification - %s' % (sqs_message['policy']['name'])
    subject = sqs_message['action'].get('subject', default_subject)
    jinja_template = jinja2.Template(subject)
    subject = jinja_template.render(
        account=sqs_message.get('account', ''),
        account_id=sqs_message.get('account_id', ''),
        event=sqs_message.get('event', None),
        action=sqs_message['action'],
        policy=sqs_message['policy'],
        region=sqs_message.get('region', '')
    )
    return subject


def setup_defaults(config):
    config.setdefault('region', 'us-east-1')
    config.setdefault('ses_region', config.get('region'))
    config.setdefault('memory', 1024)
    config.setdefault('runtime', 'python2.7')
    config.setdefault('timeout', 300)
    config.setdefault('subnets', None)
    config.setdefault('security_groups', None)
    config.setdefault('contact_tags', [])
    config.setdefault('ldap_uri', None)
    config.setdefault('ldap_bind_dn', None)
    config.setdefault('ldap_bind_user', None)
    config.setdefault('ldap_bind_password', None)
    config.setdefault('datadog_api_key', None)
    config.setdefault('slack_token', None)
    config.setdefault('slack_webhook', None)


def date_time_format(utc_str, tz_str='US/Eastern', format='%Y %b %d %H:%M %Z'):
    return parser.parse(utc_str).astimezone(gettz(tz_str)).strftime(format)


def get_date_time_delta(delta):
    return str(datetime.now().replace(tzinfo=gettz('UTC')) + timedelta(delta))


def get_date_age(date):
    return (datetime.now(tz=tzutc()) - parser.parse(date)).days


def format_struct(evt):
    return json.dumps(evt, indent=2, ensure_ascii=False)


def get_resource_tag_value(resource, k):
    for t in resource.get('Tags', []):
        if t['Key'] == k:
            return t['Value']
    return ''


def resource_format(resource, resource_type):
    if resource_type == 'ec2':
        tag_map = {t['Key']: t['Value'] for t in resource.get('Tags', ())}
        return "%s %s %s %s %s %s" % (
            resource['InstanceId'],
            resource.get('VpcId', 'NO VPC!'),
            resource['InstanceType'],
            resource.get('LaunchTime'),
            tag_map.get('Name', ''),
            resource.get('PrivateIpAddress'))
    elif resource_type == 'ami':
        return "%s %s %s" % (
            resource.get('Name'), resource['ImageId'], resource['CreationDate'])
    elif resource_type == 'sagemaker-notebook':
        return "%s" % (resource['NotebookInstanceName'])
    elif resource_type == 's3':
        return "%s" % (resource['Name'])
    elif resource_type == 'ebs':
        return "%s %s %s %s" % (
            resource['VolumeId'],
            resource['Size'],
            resource['State'],
            resource['CreateTime'])
    elif resource_type == 'rds':
        return "%s %s %s %s" % (
            resource['DBInstanceIdentifier'],
            "%s-%s" % (
                resource['Engine'], resource['EngineVersion']),
            resource['DBInstanceClass'],
            resource['AllocatedStorage'])
    elif resource_type == 'asg':
        tag_map = {t['Key']: t['Value'] for t in resource.get('Tags', ())}
        return "%s %s %s" % (
            resource['AutoScalingGroupName'],
            tag_map.get('Name', ''),
            "instances: %d" % (len(resource.get('Instances', []))))
    elif resource_type == 'elb':
        tag_map = {t['Key']: t['Value'] for t in resource.get('Tags', ())}
        if 'ProhibitedPolicies' in resource:
            return "%s %s %s %s" % (
                resource['LoadBalancerName'],
                "instances: %d" % len(resource['Instances']),
                "zones: %d" % len(resource['AvailabilityZones']),
                "prohibited_policies: %s" % ','.join(
                    resource['ProhibitedPolicies']))
        return "%s %s %s" % (
            resource['LoadBalancerName'],
            "instances: %d" % len(resource['Instances']),
            "zones: %d" % len(resource['AvailabilityZones']))
    elif resource_type == 'redshift':
        return "%s %s %s" % (
            resource['ClusterIdentifier'],
            'nodes:%d' % len(resource['ClusterNodes']),
            'encrypted:%s' % resource['Encrypted'])
    elif resource_type == 'emr':
        return "%s status:%s" % (
            resource['Id'],
            resource['Status']['State'])
    elif resource_type == 'cfn':
        return "%s" % (
            resource['StackName'])
    elif resource_type == 'launch-config':
        return "%s" % (
            resource['LaunchConfigurationName'])
    elif resource_type == 'security-group':
        name = resource.get('GroupName', '')
        for t in resource.get('Tags', ()):
            if t['Key'] == 'Name':
                name = t['Value']
        return "%s %s %s inrules: %d outrules: %d" % (
            name,
            resource['GroupId'],
            resource.get('VpcId', 'na'),
            len(resource.get('IpPermissions', ())),
            len(resource.get('IpPermissionsEgress', ())))
    elif resource_type == 'log-group':
        if 'lastWrite' in resource:
            return "name: %s last_write: %s" % (
                resource['logGroupName'],
                resource['lastWrite'])
        return "name: %s" % (resource['logGroupName'])
    elif resource_type == 'cache-cluster':
        return "name: %s created: %s status: %s" % (
            resource['CacheClusterId'],
            resource['CacheClusterCreateTime'],
            resource['CacheClusterStatus'])
    elif resource_type == 'cache-snapshot':
        cid = resource.get('CacheClusterId')
        if cid is None:
            cid = ', '.join([
                ns['CacheClusterId'] for ns in resource['NodeSnapshots']])
        return "name: %s cluster: %s source: %s" % (
            resource['SnapshotName'],
            cid,
            resource['SnapshotSource'])
    elif resource_type == 'redshift-snapshot':
        return "name: %s db: %s" % (
            resource['SnapshotIdentifier'],
            resource['DBName'])
    elif resource_type == 'ebs-snapshot':
        return "name: %s date: %s" % (
            resource['SnapshotId'],
            resource['StartTime'])
    elif resource_type == 'subnet':
        return "%s %s %s %s %s %s" % (
            resource['SubnetId'],
            resource['VpcId'],
            resource['AvailabilityZone'],
            resource['State'],
            resource['CidrBlock'],
            resource['AvailableIpAddressCount'])
    elif resource_type == 'account':
        return " %s %s" % (
            resource['account_id'],
            resource['account_name'])
    elif resource_type == 'cloudtrail':
        return " %s %s" % (
            resource['account_id'],
            resource['account_name'])
    elif resource_type == 'vpc':
        return "%s " % (
            resource['VpcId'])
    elif resource_type == 'iam-group':
        return " %s %s %s" % (
            resource['GroupName'],
            resource['Arn'],
            resource['CreateDate'])
    elif resource_type == 'rds-snapshot':
        return " %s %s %s" % (
            resource['DBSnapshotIdentifier'],
            resource['DBInstanceIdentifier'],
            resource['SnapshotCreateTime'])
    elif resource_type == 'iam-user':
        return " %s " % (
            resource['UserName'])
    elif resource_type == 'iam-role':
        return " %s %s " % (
            resource['RoleName'],
            resource['CreateDate'])
    elif resource_type == 'iam-policy':
        return " %s " % (
            resource['PolicyName'])
    elif resource_type == 'iam-profile':
        return " %s " % (
            resource['InstanceProfileId'])
    elif resource_type == 'dynamodb-table':
        return "name: %s created: %s status: %s" % (
            resource['TableName'],
            resource['CreationDateTime'],
            resource['TableStatus'])
    elif resource_type == "sqs":
        return "QueueURL: %s QueueArn: %s " % (
            resource['QueueUrl'],
            resource['QueueArn'])
    elif resource_type == "efs":
        return "name: %s  id: %s  state: %s" % (
            resource['Name'],
            resource['FileSystemId'],
            resource['LifeCycleState']
        )
    elif resource_type == "network-addr":
        return "ip: %s  id: %s  scope: %s" % (
            resource['PublicIp'],
            resource['AllocationId'],
            resource['Domain']
        )
    elif resource_type == "route-table":
        return "id: %s  vpc: %s" % (
            resource['RouteTableId'],
            resource['VpcId']
        )
    elif resource_type == "app-elb":
        return "arn: %s  zones: %s  scheme: %s" % (
            resource['LoadBalancerArn'],
            len(resource['AvailabilityZones']),
            resource['Scheme'])
    elif resource_type == "nat-gateway":
        return "id: %s  state: %s  vpc: %s" % (
            resource['NatGatewayId'],
            resource['State'],
            resource['VpcId'])
    elif resource_type == "internet-gateway":
        return "id: %s  attachments: %s" % (
            resource['InternetGatewayId'],
            len(resource['Attachments']))
    elif resource_type == 'lambda':
        return "Name: %s  RunTime: %s  \n" % (
            resource['FunctionName'],
            resource['Runtime'])
    else:
        return "%s" % format_struct(resource)


def get_provider(mailer_config):
    if mailer_config.get('queue_url', '').startswith('asq'):
        return Providers.Azure

    return Providers.AWS


def kms_decrypt(config, logger, session, encrypted_field):
    if config.get(encrypted_field):
        try:
            kms = session.client('kms')
            return kms.decrypt(
                CiphertextBlob=base64.b64decode(config[encrypted_field]))[
                    'Plaintext'].decode('utf8')
        except (TypeError, base64.binascii.Error) as e:
            logger.warning(
                "Error: %s Unable to base64 decode %s, will assume plaintext." %
                (e, encrypted_field))
        except ClientError as e:
            if e.response['Error']['Code'] != 'InvalidCiphertextException':
                raise
            logger.warning(
                "Error: %s Unable to decrypt %s with kms, will assume plaintext." %
                (e, encrypted_field))
        return config[encrypted_field]
    else:
        logger.debug("No encrypted value to decrypt.")
        return None


def decrypt(config, logger, session, encrypted_field):
    if config.get(encrypted_field):
        provider = get_provider(config)
        if provider == Providers.Azure:
            return config[encrypted_field]
        elif provider == Providers.AWS:
            return kms_decrypt(config, logger, session, encrypted_field)
        else:
            raise Exception("Unknown provider")
    else:
        logger.debug("No encrypted value to decrypt.")
        return None


# https://docs.aws.amazon.com/awscloudtrail/latest/userguide/cloudtrail-event-reference-user-identity.html
def get_aws_username_from_event(logger, event):
    if event is None:
        return None
    identity = event.get('detail', {}).get('userIdentity', {})
    if not identity:
        logger.warning("Could not get recipient from event \n %s" % (
            format_struct(event)))
        return None
    if identity['type'] == 'AssumedRole':
        logger.debug(
            'In some cases there is no ldap uid is associated with AssumedRole: %s',
            identity['arn'])
        logger.debug(
            'We will try to assume that identity is in the AssumedRoleSessionName')
        user = identity['arn'].rsplit('/', 1)[-1]
        if user is None or user.startswith('i-') or user.startswith('awslambda'):
            return None
        if ':' in user:
            user = user.split(':', 1)[-1]
        return user
    if identity['type'] == 'IAMUser' or identity['type'] == 'WebIdentityUser':
        return identity['userName']
    if identity['type'] == 'Root':
        return None
    # this conditional is left here as a last resort, it should
    # be better documented with an example UserIdentity json
    if ':' in identity['principalId']:
        user_id = identity['principalId'].split(':', 1)[-1]
    else:
        user_id = identity['principalId']
    return user_id



# Copyright 2016-2017 Capital One Services, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

import copy
import json
import logging
import os

try:
    from c7n_azure.function_package import FunctionPackage
    from c7n_azure.functionapp_utils import FunctionAppUtilities
    from c7n_azure.policy import AzureFunctionMode
    from c7n_azure.session import Session
    from c7n_azure.utils import StringUtils
    from c7n.utils import local_session
except ImportError:
    FunctionPackage = None
    pass


def build_function_package(config, function_name):
    schedule = config.get('function_schedule', '0 */10 * * * *')

    # Build package
    package = FunctionPackage(
        function_name,
        os.path.join(os.path.dirname(__file__), 'function.py'))

    package.build(None,
                  modules=['c7n', 'c7n-azure', 'c7n-mailer', 'applicationinsights'],
                  non_binary_packages=['pyyaml', 'pycparser', 'tabulate', 'jmespath',
                                       'datadog', 'MarkupSafe', 'simplejson', 'pyrsistent'],
                  excluded_packages=['azure-cli-core', 'distlib', 'future', 'futures'])

    package.pkg.add_contents(
        function_name + '/function.json',
        contents=package.get_function_config({'mode':
                                              {'type': 'azure-periodic',
                                               'schedule': schedule}}))

    # Add mail templates
    for d in set(config['templates_folders']):
        if not os.path.exists(d):
            continue
        for t in [f for f in os.listdir(d) if os.path.splitext(f)[1] == '.j2']:
            with open(os.path.join(d, t)) as fh:
                package.pkg.add_contents(function_name + '/msg-templates/%s' % t, fh.read())

    function_config = copy.deepcopy(config)
    function_config['templates_folders'] = [function_name + '/msg-templates/']
    package.pkg.add_contents(
        function_name + '/config.json',
        contents=json.dumps(function_config))

    package.close()
    return package


def provision(config):
    log = logging.getLogger('c7n_mailer.azure.deploy')

    function_name = config.get('function_name', 'mailer')
    function_properties = config.get('function_properties', {})

    # service plan is parse first, because its location might be shared with storage & insights
    service_plan = AzureFunctionMode.extract_properties(function_properties,
                                                'servicePlan',
                                                {
                                                    'name': 'cloud-custodian',
                                                    'location': 'eastus',
                                                    'resource_group_name': 'cloud-custodian',
                                                    'sku_tier': 'Dynamic',  # consumption plan
                                                    'sku_name': 'Y1'
                                                })

    location = service_plan.get('location', 'eastus')
    rg_name = service_plan['resource_group_name']

    sub_id = local_session(Session).get_subscription_id()
    suffix = StringUtils.naming_hash(rg_name + sub_id)

    storage_account = AzureFunctionMode.extract_properties(function_properties,
                                                    'storageAccount',
                                                    {'name': 'mailerstorage' + suffix,
                                                     'location': location,
                                                     'resource_group_name': rg_name})

    app_insights = AzureFunctionMode.extract_properties(function_properties,
                                                    'appInsights',
                                                    {'name': service_plan['name'],
                                                     'location': location,
                                                     'resource_group_name': rg_name})

    function_app_name = FunctionAppUtilities.get_function_name(
        '-'.join([service_plan['name'], function_name]), suffix)
    FunctionAppUtilities.validate_function_name(function_app_name)

    params = FunctionAppUtilities.FunctionAppInfrastructureParameters(
        app_insights=app_insights,
        service_plan=service_plan,
        storage_account=storage_account,
        function_app_resource_group_name=service_plan['resource_group_name'],
        function_app_name=function_app_name)

    FunctionAppUtilities.deploy_function_app(params)

    log.info("Building function package for %s" % function_app_name)
    package = build_function_package(config, function_name)

    log.info("Function package built, size is %dMB" % (package.pkg.size / (1024 * 1024)))

    FunctionAppUtilities.publish_functions_package(params, package)
